"""Generate simple tables in terminals from a nested list of strings.

Use SingleTable or DoubleTable instead of AsciiTable for box-drawing characters.

https://github.com/Robpol86/terminaltables
https://pypi.python.org/pypi/terminaltables
"""

from terminaltables.tables import AsciiTable, DoubleTable, SingleTable


assert AsciiTable
assert DoubleTable
assert SingleTable



"""Functions that handle alignment, padding, widths, etc."""

import unicodedata


def string_width(string):
    """Get the visible width of a unicode string.

    Some CJK unicode characters are more than one byte unlike ASCII and latin unicode characters.

    From: https://github.com/Robpol86/terminaltables/pull/9

    :param str string: String to measure.

    :return: String's width.
    :rtype: int
    """
    # Colorclass instance.
    if hasattr(string, 'value_no_colors'):
        string = string.value_no_colors

    # Convert to unicode.
    try:
        decoded = string.decode('u8')
    except (AttributeError, UnicodeEncodeError):
        decoded = string

    width = 0
    for char in decoded:
        if unicodedata.east_asian_width(char) in ('F', 'W'):
            width += 2
        else:
            width += 1

    return width


def align_and_pad_cell(string, align, size_dims):
    """Align a string with center/rjust/ljust and adds additional padding.

    :param str string: Input string to operate on.
    :param str align: 'left', 'right', or 'center'.
    :param iter size_dims: Size and dimensions. A 4-item tuple of integers representing width, height, lpad, and rpad.

    :return: Modified string.
    :rtype: str
    """
    width, height, lpad, rpad = size_dims

    # Handle trailing newlines or empty strings, str.splitlines() does not satisfy.
    lines = string.splitlines() or ['']
    if string.endswith('\n'):
        lines.append('')

    # Align.
    if align == 'center':
        method = 'center'
    elif align == 'right':
        method = 'rjust'
    else:
        method = 'ljust'
    aligned = '\n'.join(getattr(l, method)(width + len(l) - string_width(l)) for l in lines)

    # Pad.
    padded = '\n'.join((' ' * lpad) + l + (' ' * rpad) for l in aligned.splitlines() or [''])

    # Increase height.
    additional_padding = height - 1 - padded.count('\n')
    if additional_padding > 0:
        padded += ('\n{0}'.format(' ' * (width + lpad + rpad))) * additional_padding

    return padded


def column_widths(table_data):
    """Get maximum widths of each column in the table.

    :param iter table_data: List of list of strings. The unpadded table data.

    :return: Column widths.
    :rtype: list
    """
    if not table_data:
        return list()

    number_of_columns = max(len(r) for r in table_data)
    widths = [0] * number_of_columns

    for row in table_data:
        for i in range(len(row)):
            if not row[i]:
                continue
            widths[i] = max(widths[i], string_width(max(row[i].splitlines(), key=len)))

    return widths



"""Main table class."""

import re

from terminaltables.terminal_io import terminal_size
from terminaltables.width_and_alignment import align_and_pad_cell, column_widths, string_width


def join_row(row, left, middle, right):
    """Convert a row (list of strings) into a joined string with left and right borders. Supports multi-lines.

    :param iter row: List of strings representing one row.
    :param str left: Left border.
    :param str middle: Column separator.
    :param str right: Right border.

    :return: String representation of a row.
    :rtype: str
    """
    if not row:
        return left + right

    if not any('\n' in c for c in row):
        return left + middle.join(row) + right

    # Split cells in the row by newlines. This creates new rows.
    split_cells = [(c.splitlines() or ['']) + ([''] if c.endswith('\n') else []) for c in row]
    height = len(split_cells[0])

    # Merge rows into strings.
    converted_rows = list()
    for row_number in range(height):
        converted_rows.append(left + middle.join([c[row_number] for c in split_cells]) + right)
    return '\n'.join(converted_rows)


class BaseTable(object):
    """Base table class."""

    CHAR_CORNER_LOWER_LEFT = ''
    CHAR_CORNER_LOWER_RIGHT = ''
    CHAR_CORNER_UPPER_LEFT = ''
    CHAR_CORNER_UPPER_RIGHT = ''
    CHAR_HORIZONTAL = ''
    CHAR_INTERSECT_BOTTOM = ''
    CHAR_INTERSECT_CENTER = ''
    CHAR_INTERSECT_LEFT = ''
    CHAR_INTERSECT_RIGHT = ''
    CHAR_INTERSECT_TOP = ''
    CHAR_VERTICAL = ''

    def __init__(self, table_data, title=None):
        """Constructor.

        :param iter table_data: List (empty or list of lists of strings) representing the table.
        :param str title: Optional title to show within the top border of the table.
        """
        self.table_data = table_data
        self.title = title

        self.inner_column_border = True
        self.inner_heading_row_border = True
        self.inner_footing_row_border = False
        self.inner_row_border = False
        self.justify_columns = dict()  # {0: 'right', 1: 'left', 2: 'center'}
        self.outer_border = True
        self.padding_left = 1
        self.padding_right = 1

    def column_max_width(self, column_number):
        """Return the maximum width of a column based on the current terminal width.

        :param int column_number: The column number to query.

        :return: The max width of the column.
        :rtype: int
        """
        widths = self.column_widths
        borders_padding = (len(widths) * self.padding_left) + (len(widths) * self.padding_right)
        if self.outer_border:
            borders_padding += 2
        if self.inner_column_border and widths:
            borders_padding += len(widths) - 1
        other_column_widths = sum(widths) - widths[column_number]
        return terminal_size()[0] - other_column_widths - borders_padding

    @property
    def column_widths(self):
        """Return a list of integers representing the widths of each table column without padding."""
        if not self.table_data:
            return list()
        return column_widths(self.table_data)

    @property
    def ok(self):  # Too late to change API. # pylint: disable=invalid-name
        """Return True if the table fits within the terminal width, False if the table breaks."""
        return self.table_width <= terminal_size()[0]

    @property
    def padded_table_data(self):
        """Return a list of lists of strings. It's self.table_data but with the cells padded with spaces and newlines.

        Most of the work in this class is done here.
        """
        if not self.table_data:
            return list()

        # Set all rows to the same number of columns.
        max_columns = max(len(r) for r in self.table_data)
        new_table_data = [r + [''] * (max_columns - len(r)) for r in self.table_data]

        # Pad strings in each cell, and apply text-align/justification.
        widths = self.column_widths
        for row in new_table_data:
            height = max([c.count('\n') for c in row] or [0]) + 1
            for i in range(len(row)):
                align = self.justify_columns.get(i, 'left')
                cell = align_and_pad_cell(row[i], align, (widths[i], height, self.padding_left, self.padding_right))
                row[i] = cell

        return new_table_data

    @property
    def table(self):
        """Return a large string of the entire table ready to be printed to the terminal."""
        padded_table_data = self.padded_table_data
        widths = [c + self.padding_left + self.padding_right for c in self.column_widths]
        final_table_data = list()

        # Append top border.
        max_title = sum(widths) + ((len(widths) - 1) if self.inner_column_border else 0)
        if self.outer_border and self.title and string_width(self.title) <= max_title:
            pseudo_row = join_row(
                ['h' * w for w in widths],
                'l', 't' if self.inner_column_border else '',
                'r'
            )
            pseudo_row_key = dict(h=self.CHAR_HORIZONTAL, l=self.CHAR_CORNER_UPPER_LEFT, t=self.CHAR_INTERSECT_TOP,
                                  r=self.CHAR_CORNER_UPPER_RIGHT)
            pseudo_row_re = re.compile('({0})'.format('|'.join(pseudo_row_key.keys())))
            substitute = lambda s: pseudo_row_re.sub(lambda x: pseudo_row_key[x.string[x.start():x.end()]], s)
            row = substitute(pseudo_row[:1]) + self.title + substitute(pseudo_row[1 + string_width(self.title):])
            final_table_data.append(row)
        elif self.outer_border:
            row = join_row(
                [self.CHAR_HORIZONTAL * w for w in widths],
                self.CHAR_CORNER_UPPER_LEFT,
                self.CHAR_INTERSECT_TOP if self.inner_column_border else '',
                self.CHAR_CORNER_UPPER_RIGHT
            )
            final_table_data.append(row)

        # Build table body.
        indexes = range(len(padded_table_data))
        for i in indexes:
            row = join_row(
                padded_table_data[i],
                self.CHAR_VERTICAL if self.outer_border else '',
                self.CHAR_VERTICAL if self.inner_column_border else '',
                self.CHAR_VERTICAL if self.outer_border else ''
            )
            final_table_data.append(row)

            # Insert row separator.
            if i == indexes[-1]:
                continue
            if self.inner_row_border or (self.inner_heading_row_border and i == 0):
                row = join_row(
                    [self.CHAR_HORIZONTAL * w for w in widths],
                    self.CHAR_INTERSECT_LEFT if self.outer_border else '',
                    self.CHAR_INTERSECT_CENTER if self.inner_column_border else '',
                    self.CHAR_INTERSECT_RIGHT if self.outer_border else ''
                )
                final_table_data.append(row)

            if i == indexes[-2] and self.inner_footing_row_border:
                row = join_row(
                    [self.CHAR_HORIZONTAL * w for w in widths],
                    self.CHAR_INTERSECT_LEFT if self.outer_border else '',
                    self.CHAR_INTERSECT_CENTER if self.inner_column_border else '',
                    self.CHAR_INTERSECT_RIGHT if self.outer_border else ''
                )
                final_table_data.append(row)

        # Append bottom border.
        if self.outer_border:
            row = join_row(
                [self.CHAR_HORIZONTAL * w for w in widths],
                self.CHAR_CORNER_LOWER_LEFT,
                self.CHAR_INTERSECT_BOTTOM if self.inner_column_border else '',
                self.CHAR_CORNER_LOWER_RIGHT
            )
            final_table_data.append(row)

        return '\n'.join(final_table_data)

    @property
    def table_width(self):
        """Return the width of the table including padding and borders."""
        widths = self.column_widths
        borders_padding = (len(widths) * self.padding_left) + (len(widths) * self.padding_right)
        if self.outer_border:
            borders_padding += 2
        if self.inner_column_border and widths:
            borders_padding += len(widths) - 1
        return sum(widths) + borders_padding



"""Get info about the current terminal window/screen buffer."""

import ctypes
import struct
import sys

try:
    import fcntl
    import termios
except ImportError:
    fcntl = None
    termios = None

DEFAULT_HEIGHT = 24
DEFAULT_WIDTH = 80


def terminal_size():
    """Get the width and height of the terminal.

    http://code.activestate.com/recipes/440694-determine-size-of-console-window-on-windows/
    http://stackoverflow.com/questions/17993814/why-the-irrelevant-code-made-a-difference

    :return: Width (number of characters) and height (number of lines) of the terminal.
    :rtype: tuple
    """
    if hasattr(ctypes, 'windll'):
        # Only works on Microsoft Windows platforms.
        string_buffer = ctypes.create_string_buffer(22)  # To be written to by GetConsoleScreenBufferInfo.
        ctypes.windll.kernel32.GetConsoleScreenBufferInfo(ctypes.windll.kernel32.GetStdHandle(-11), string_buffer)
        left, top, right, bottom = struct.unpack('hhhhHhhhhhh', string_buffer.raw)[5:-2]
        width, height = right - left, bottom - top
        if width < 1 or height < 1:
            return DEFAULT_WIDTH, DEFAULT_HEIGHT
        return width, height

    try:
        device = fcntl.ioctl(0, termios.TIOCGWINSZ, '\0' * 8)
    except IOError:
        return DEFAULT_WIDTH, DEFAULT_HEIGHT
    height, width = struct.unpack('hhhh', device)[:2]
    return width, height


def set_terminal_title(title):
    """Set the terminal title.

    :param str title: The title to set.
    """
    if hasattr(ctypes, 'windll'):
        if sys.version_info[0] == 3:
            func = ctypes.windll.kernel32.SetConsoleTitleW  # Unicode.
        else:
            func = ctypes.windll.kernel32.SetConsoleTitleA  # Ascii.
        return func(title)
    sys.stdout.write('\033]0;{0}\007'.format(title))



#!/usr/bin/env python
"""Setup script for the project."""

from __future__ import print_function

import codecs
import os
import sys
from distutils import core

from setuptools import setup


class DumpRequirements(core.Command):
    """Write requirements.txt file for tox."""

    description = 'Write requirements.txt file for tox.'
    user_options = [('outfile=', 'o', 'File path to write to or stderr if "-".')]

    def initialize_options(self):
        """Required by distutils. Using setattr() instead of self. due to PyCharm inspections."""
        setattr(self, 'outfile', None)

    def finalize_options(self):
        """Required by distutils."""
        outfile = os.path.realpath(getattr(self, 'outfile') or 'requirements.txt')
        setattr(self, 'outfile', outfile)

    def run(self):
        """Write to file or standard err (distutils pollutes stdout)."""
        dist = getattr(core, '_setup_distribution')
        requirements = '\n'.join(sorted(set(dist.install_requires + dist.tests_require)))
        outfile = getattr(self, 'outfile')
        if outfile.endswith('-'):
            return print(requirements, file=sys.stderr)
        try:
            with open(outfile, 'w') as handle:
                handle.write(requirements)
        except IOError:
            raise SystemExit('Unable to write to {0}'.format(outfile))
        print('Wrote to {0}'.format(outfile))


def safe_read(path):
    """Try to read file or return empty string if failed.

    :param str path: Relative file path to read.

    :return: File contents.
    :rtype: str
    """
    abspath, file_handle = os.path.join(os.path.abspath(os.path.dirname(__file__)), path), None
    try:
        file_handle = codecs.open(abspath, encoding='utf-8')
        return file_handle.read(131072)
    except IOError:
        return ''
    finally:
        getattr(file_handle, 'close', lambda: None)()


setup(
    author='@Robpol86',
    author_email='robpol86@gmail.com',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Environment :: Console',
        'Environment :: MacOS X',
        'Environment :: Win32 (MS Windows)',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Topic :: Software Development :: Libraries',
        'Topic :: Terminals',
        'Topic :: Text Processing :: Markup',
    ],
    cmdclass=dict(requirements=DumpRequirements),
    description='Generate simple tables in terminals from a nested list of strings.',
    install_requires=[],
    keywords='Shell Bash ANSI ASCII terminal tables',
    license='MIT',
    long_description=safe_read('README.rst'),
    name='terminaltables',
    packages=['terminaltables'],
    tests_require=['colorclass', 'pytest-cov'],
    url='https://github.com/Robpol86/terminaltables',
    version='2.0.0',
    zip_safe=True,
)



"""User-facing tables defined here."""

import os

from terminaltables.base_table import BaseTable


class AsciiTable(BaseTable):
    """Draw a table using regular ASCII characters, such as `+`, `|`, and `-`."""

    CHAR_CORNER_LOWER_LEFT = '+'
    CHAR_CORNER_LOWER_RIGHT = '+'
    CHAR_CORNER_UPPER_LEFT = '+'
    CHAR_CORNER_UPPER_RIGHT = '+'
    CHAR_HORIZONTAL = '-'
    CHAR_INTERSECT_BOTTOM = '+'
    CHAR_INTERSECT_CENTER = '+'
    CHAR_INTERSECT_LEFT = '+'
    CHAR_INTERSECT_RIGHT = '+'
    CHAR_INTERSECT_TOP = '+'
    CHAR_VERTICAL = '|'


class UnixTable(BaseTable):
    """Draw a table using box-drawing characters on Unix platforms. Table borders won't have any gaps between lines.

    Similar to the tables shown on PC BIOS boot messages, but not double-lined.
    """

    CHAR_CORNER_LOWER_LEFT = '\033(0\x6d\033(B'
    CHAR_CORNER_LOWER_RIGHT = '\033(0\x6a\033(B'
    CHAR_CORNER_UPPER_LEFT = '\033(0\x6c\033(B'
    CHAR_CORNER_UPPER_RIGHT = '\033(0\x6b\033(B'
    CHAR_HORIZONTAL = '\033(0\x71\033(B'
    CHAR_INTERSECT_BOTTOM = '\033(0\x76\033(B'
    CHAR_INTERSECT_CENTER = '\033(0\x6e\033(B'
    CHAR_INTERSECT_LEFT = '\033(0\x74\033(B'
    CHAR_INTERSECT_RIGHT = '\033(0\x75\033(B'
    CHAR_INTERSECT_TOP = '\033(0\x77\033(B'
    CHAR_VERTICAL = '\033(0\x78\033(B'

    @property
    def table(self):
        """Return a large string of the entire table ready to be printed to the terminal."""
        ascii_table = super(UnixTable, self).table
        optimized = ascii_table.replace('\033(B\033(0', '')
        return optimized


class WindowsTable(BaseTable):
    """Draw a table using box-drawing characters on Windows platforms. This uses Code Page 437. Single-line borders.

    From: http://en.wikipedia.org/wiki/Code_page_437#Characters
    """

    CHAR_CORNER_LOWER_LEFT = b'\xc0'.decode('ibm437')
    CHAR_CORNER_LOWER_RIGHT = b'\xd9'.decode('ibm437')
    CHAR_CORNER_UPPER_LEFT = b'\xda'.decode('ibm437')
    CHAR_CORNER_UPPER_RIGHT = b'\xbf'.decode('ibm437')
    CHAR_HORIZONTAL = b'\xc4'.decode('ibm437')
    CHAR_INTERSECT_BOTTOM = b'\xc1'.decode('ibm437')
    CHAR_INTERSECT_CENTER = b'\xc5'.decode('ibm437')
    CHAR_INTERSECT_LEFT = b'\xc3'.decode('ibm437')
    CHAR_INTERSECT_RIGHT = b'\xb4'.decode('ibm437')
    CHAR_INTERSECT_TOP = b'\xc2'.decode('ibm437')
    CHAR_VERTICAL = b'\xb3'.decode('ibm437')


class WindowsTableDouble(BaseTable):
    """Draw a table using box-drawing characters on Windows platforms. This uses Code Page 437. Double-line borders."""

    CHAR_CORNER_LOWER_LEFT = b'\xc8'.decode('ibm437')
    CHAR_CORNER_LOWER_RIGHT = b'\xbc'.decode('ibm437')
    CHAR_CORNER_UPPER_LEFT = b'\xc9'.decode('ibm437')
    CHAR_CORNER_UPPER_RIGHT = b'\xbb'.decode('ibm437')
    CHAR_HORIZONTAL = b'\xcd'.decode('ibm437')
    CHAR_INTERSECT_BOTTOM = b'\xca'.decode('ibm437')
    CHAR_INTERSECT_CENTER = b'\xce'.decode('ibm437')
    CHAR_INTERSECT_LEFT = b'\xcc'.decode('ibm437')
    CHAR_INTERSECT_RIGHT = b'\xb9'.decode('ibm437')
    CHAR_INTERSECT_TOP = b'\xcb'.decode('ibm437')
    CHAR_VERTICAL = b'\xba'.decode('ibm437')


class SingleTable(WindowsTable if os.name == 'nt' else UnixTable):
    """Cross-platform table with single-line box-drawing characters."""

    pass


class DoubleTable(WindowsTableDouble):
    """Cross-platform table with box-drawing characters. On Windows it's double borders, on Linux/OSX it's unicode."""

    pass
